/*
    Copyright (C) 2008    Dmitry Ivanov <vonami@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "feedlist.h"
#include "feedlist_p.h"
#include "resource.h"
#include "resourcemanager.h"
#include "persistentfeed.h"
#include "feedcollection.h"

#include <akonadi/collectionfetchjob.h>

#include <KDebug>
#include <KJob>

#include <QtCore/QStringList>

#include <algorithm>
#include <memory>

#include <cassert>

using Akonadi::Collection;
using Akonadi::CollectionFetchJob;

using namespace KRss;
using namespace boost;


// private implementation
Feed* FeedListPrivate::appendPersistentFeed( const FeedCollection &collection )
{
    feedsMap[ collection.resource() ].append( collection.id() );
    Feed *feed = new PersistentFeed( collection, resources.value( collection.resource() ) );
    feeds.insert( collection.id(), feed );

    QObject::connect( feed, SIGNAL( unreadCountChanged( KRss::Feed*, int ) ),
                      q, SIGNAL( unreadCountChanged( KRss::Feed*, int ) ) );
    QObject::connect( feed, SIGNAL( totalCountChanged( KRss::Feed*, int ) ),
                      q, SIGNAL( totalCountChanged( KRss::Feed*, int ) ) );

    QObject::connect( feed, SIGNAL( fetchStarted( KRss::Feed* ) ), q, SIGNAL( fetchStarted( KRss::Feed* ) ) );
    QObject::connect( feed, SIGNAL( fetchPercent( KRss::Feed*, uint ) ), q, SIGNAL( fetchPercent( KRss::Feed*, uint ) ) );
    QObject::connect( feed, SIGNAL( fetchFinished( KRss::Feed* ) ), q, SIGNAL( fetchFinished( KRss::Feed* ) ) );
    QObject::connect( feed, SIGNAL( fetchFailed( KRss::Feed*, QString ) ),
                      q, SIGNAL( fetchFailed( KRss::Feed*, QString ) ) );
    QObject::connect( feed, SIGNAL( changed( KRss::Feed* ) ), q, SIGNAL( feedChanged( KRss::Feed* ) ) );
    QObject::connect( feed, SIGNAL( removed( KRss::Feed* ) ), q, SLOT( slotFeedRemoved( KRss::Feed* ) ) );

    return feed;
}

void FeedListPrivate::dump()
{
    QHashIterator<QString, QList<Feed::Id> > it_res( feedsMap );
    while( it_res.hasNext() ) {
        it_res.next();
        kDebug() << "Resource:" << it_res.key() << ", feeds:" << it_res.value();
    }

    QHashIterator<Feed::Id, Feed* > it_feed( feeds );
    while( it_feed.hasNext() ) {
        it_feed.next();
        kDebug() << "Feed id:" << it_feed.key() << ", title:" << it_feed.value()->title();
    }
}

void FeedListPrivate::slotFeedAdded( Feed::Id id )
{
    CollectionFetchJob *job = new CollectionFetchJob( Collection( id ), CollectionFetchJob::Base, q );
    QObject::connect( job, SIGNAL( result( KJob* ) ), q, SLOT( slotCollectionLoadDone( KJob* ) ) );
}

void FeedListPrivate::slotFeedRemoved( Feed *feed )
{
    Q_UNUSED( feed )
    //TODO: implement
}

void FeedListPrivate::slotCollectionLoadDone( KJob *job )
{
    if ( job->error() ) {
        kWarning() << "Failed to load collection from a resource";
        kWarning() << job->errorString();
        return;
    }

    Collection col = static_cast<CollectionFetchJob*>( job )->collections().first();
    Feed *feed = appendPersistentFeed( col );
    emit q->feedAdded( feed );
    dump();
}

// public interface implementation
FeedList::FeedList( const QStringList &resourceIdentifiers, QObject *parent )
    : QObject( parent ), d ( new FeedListPrivate( this ) )
{
    Q_FOREACH( const QString &resourceIdentifier, resourceIdentifiers ) {
        const Resource *resource = ResourceManager::self()->resource( resourceIdentifier );
        if ( !resource ) {
            kDebug() << "Resource" << resourceIdentifier << " is not valid";
            continue;
        }

        d->resources.insert( resourceIdentifier, resource );
        connect( resource, SIGNAL( feedAdded( KRss::Feed::Id ) ), this, SLOT( slotFeedAdded( KRss::Feed::Id ) ) );
    }
}

FeedList::~FeedList()
{
    delete d;
}

QList<Feed*> FeedList::feeds() const
{
    return d->feeds.values();
}

#if 0
QList<const Feed*> FeedList::feeds() const
{
    QList<const Feed*> l;
    Q_FOREACH( const Feed* const i, d->feeds.values() )
        l.append( i );
    return l;
}
#endif

QList<Feed::Id> FeedList::feedIds() const
{
    return d->feeds.keys();
}

QList<TagId> FeedList::tags() const
{
    QList<TagId> allTags;
    Q_FOREACH( const Feed * const feed, d->feeds.values() )
        allTags << feed->tags();

    std::sort( allTags.begin(), allTags.end() );
    allTags.erase( std::unique( allTags.begin(), allTags.end() ), allTags.end() );
    return allTags;
}

bool FeedList::isEmpty() const
{
    return d->feeds.isEmpty();
}

Feed* FeedList::findById( const Feed::Id& id )
{
    return d->feeds.value( id );
}

const Feed* FeedList::findById( const Feed::Id& id ) const
{
    return d->feeds.value( id );
}

QList<Feed*> FeedList::findByTitle( const QString &title )
{
    QList<Feed*> feeds;

    Q_FOREACH( Feed *feed, d->feeds.values() ) {
        if ( feed->title() == title )
            feeds << feed;
    }

    return feeds;
}

QList<const Feed*> FeedList::findByTitle( const QString &title ) const
{
    QList<const Feed*> feeds;

    Q_FOREACH( const Feed *feed, d->feeds.values() ) {
        if ( feed->title() == title )
            feeds << feed;
    }

    return feeds;
}

void FeedList::fetch( Feed::Id id )
{
    Feed *feed = d->feeds.value( id );
    if ( feed )
        feed->fetch();
}

class KRss::RetrieveFeedListJobPrivate
{
    RetrieveFeedListJob* const q;
public:
    explicit RetrieveFeedListJobPrivate( RetrieveFeedListJob* qq );
    void loadDone( KJob* job );
    void doStart();

public:
    shared_ptr<FeedList> feedList;
    int pendingFetchJobs;
    QStringList resourceIdentifiers;
};

RetrieveFeedListJobPrivate::RetrieveFeedListJobPrivate( RetrieveFeedListJob* qq )
    : q( qq )
    , feedList()
    , pendingFetchJobs( 0 )
    , resourceIdentifiers()
{
}

RetrieveFeedListJob::RetrieveFeedListJob( QObject* parent ) : KJob( parent ), d( new RetrieveFeedListJobPrivate( this ) )
{
}

RetrieveFeedListJob::~RetrieveFeedListJob()
{
    delete d;
}

void RetrieveFeedListJob::setResourceIdentifiers( const QStringList& resourceIdentifiers )
{
    d->resourceIdentifiers = resourceIdentifiers;
}

QStringList RetrieveFeedListJob::resourceIdentifiers() const
{
    return d->resourceIdentifiers;
}

void RetrieveFeedListJobPrivate::loadDone( KJob* job_ )
{
    CollectionFetchJob* const job = qobject_cast<CollectionFetchJob*>( job_ );
    assert( job );
    assert( pendingFetchJobs > 0 );
    --pendingFetchJobs;

    if ( job->error() ) {
        q->setError( RetrieveFeedListJob::CouldNotRetrieveFeedListError );
        q->setErrorText( job->errorString() );
        q->emitResult();
        return;
    }

    const QList<Collection> cols = job->collections();
    QList<KUrl> uris;
    Q_FOREACH( const Collection &col, cols ) {
        // accept only 'first generation' child collections
        if ( col.parent() != Collection::root().id() ) {
            feedList->d->appendPersistentFeed( col );
            uris.append( col.url() );
        }
    }

    if ( pendingFetchJobs > 0 )
        return;

    q->emitResult();
}

shared_ptr<FeedList> RetrieveFeedListJob::feedList() const
{
    return d->feedList;
}

void RetrieveFeedListJob::start()
{
    QMetaObject::invokeMethod( this, "doStart", Qt::QueuedConnection );
}

void RetrieveFeedListJobPrivate::doStart()
{
    feedList.reset( new FeedList( resourceIdentifiers ) );
    if ( resourceIdentifiers.isEmpty() ) {
        q->emitResult();
        return;
    }

    Q_FOREACH( const QString &resourceId, resourceIdentifiers ) {
        std::auto_ptr<CollectionFetchJob> job( new CollectionFetchJob( Collection::root(), CollectionFetchJob::Recursive ) );
        job->includeUnsubscribed( true );
        job->setResource( resourceId );
        q->connect( job.get(), SIGNAL(result(KJob*)), q, SLOT(loadDone(KJob*)) );
        ++pendingFetchJobs;
        job.release()->start();
    }
}

#include "feedlist.moc"
