/*
    Copyright (C) 2009    Frank Osterfeld <osterfeld@kde.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "itemcollection.h"
#include "feed.h"
#include "item.h"

#include <boost/bind.hpp>
#include <algorithm>

using namespace boost;
using namespace KRss;

namespace {
    static bool lessByItemId( const Item& lhs, const Item& rhs ) {
        return lhs.id() < rhs.id();
    }

    static bool sameItemId( const Item& lhs, const Item& rhs ) {
        return lhs.id() == rhs.id();
    }
}

ItemCollection::Listener::~Listener() {}

class ItemCollection::Private {
    ItemCollection* const q;
public:
    explicit Private( const QList<Item>& list, ItemCollection* qq );

    void addItems( const QList<Item>& list );
    void updateItems( const QList<Item>& list );
    void removeItems( const QList<Item>& list );

    void itemsAdded( KRss::Feed*, const QList<KRss::Item>& list ) {
        addItems( list );
    }

    void itemsChanged( KRss::Feed*, const QList<KRss::Item>& list ) {
        updateItems( list );
    }

    void itemsRemoved( KRss::Feed*, const QList<KRss::Item>& list ) {
        removeItems( list );
    }

    QList<Item> items;
    QList<ItemCollection::Listener*> listeners;
};

ItemCollection::Private::Private( const QList<Item>& list, ItemCollection* qq )
    : q( qq )
    , items( list ) {
    std::sort( items.begin(), items.end(), lessByItemId );
}

void ItemCollection::Private::addItems( const QList<Item>& list_ ) {
    if ( list_.isEmpty() )
        return;
    QList<Item> list( list_ );
    std::sort( list.begin(), list.end(), lessByItemId );
    QList<Item>::iterator from = items.begin();
    Q_FOREACH( const Item& i, list ) {
        QList<Item>::iterator it = std::lower_bound( from, items.end(), i, lessByItemId );
        const int pos = std::distance( items.begin(), it );
        if ( it != items.end() && sameItemId( *it, i ) ) {
            *it = i;
            Q_FOREACH( ItemCollection::Listener* const j, listeners )
                j->update( pos );
            from = it;
        } else {
            Q_FOREACH( ItemCollection::Listener* const j, listeners )
                j->prepareInsert( pos );
            from = items.insert( it, i );
            Q_FOREACH( ItemCollection::Listener* const j, listeners )
                j->finishInsert( pos );
        }
    }
}

void ItemCollection::Private::updateItems( const QList<Item>& list_ ) {
    QList<Item> list( list_ );
    std::sort( list.begin(), list.end(), lessByItemId );
    QList<Item>::iterator from = items.begin();
    Q_FOREACH( const Item & i, list ) {
        const QList<Item>::iterator it = qBinaryFind( from, items.end(), i, lessByItemId );
        if ( it == items.end() )
            continue;
        *it = i;
        from = it;
        Q_FOREACH( ItemCollection::Listener* const j, listeners )
            j->update( std::distance( items.begin(), it ) );
    }
}

void ItemCollection::Private::removeItems( const QList<Item>& list_ ) {
    QList<Item> list( list_ );
    std::sort( list.begin(), list.end(), lessByItemId );
    QList<Item>::iterator newEnd = items.end();
    Q_FOREACH( const Item & i, list ) {
        QList<Item>::iterator it = qBinaryFind( items.begin(), items.end(), i, lessByItemId );
        if ( it == items.end() )
            continue;
        const int pos = std::distance( items.begin(), it );
        Q_FOREACH( ItemCollection::Listener* const j, listeners )
            j->prepareRemove( pos );
        items.erase( it );
        Q_FOREACH( ItemCollection::Listener* const j, listeners )
            j->finishRemove( pos );
    }
}

ItemCollection::ItemCollection( const QList<Item>& items, QObject* parent ) : QObject( parent ), d( new Private( items, this ) ) {
}

ItemCollection::~ItemCollection() {
    delete d;
}

const QList<Item>& ItemCollection::items() const {
    return d->items;
}

void ItemCollection::connectToFeed( const Feed* feed ) {
    connect( feed, SIGNAL(itemsAdded(KRss::Feed*,QList<KRss::Item>)),
            this, SLOT(itemsAdded(KRss::Feed*,QList<KRss::Item>)) );
    connect( feed, SIGNAL(itemsChanged(KRss::Feed*,QList<KRss::Item>)),
                this, SLOT(itemsChanged(KRss::Feed*,QList<KRss::Item>)) );
    connect( feed, SIGNAL(itemsRemoved(KRss::Feed*,QList<KRss::Item>)),
                this, SLOT(itemsRemoved(KRss::Feed*,QList<KRss::Item>)) );
}

void ItemCollection::addListener( ItemCollection::Listener* listener ) {
    if ( d->listeners.contains( listener ) )
        return;
    d->listeners.append( listener );
}

void ItemCollection::removeListener( ItemCollection::Listener* listener ) {
    d->listeners.removeAll( listener );
}

#include "itemcollection.moc"

