/*
    Copyright (C) 2008    Dmitry Ivanov <vonami@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "itemstatusmodifyjob.h"
#include "itemjobs.h"
#include "feed.h"

#include <KDebug>

#include <QtCore/QTimer>

using namespace KRss;

namespace KRss {

class ItemStatusModifyJobPrivate
{
public:

    explicit ItemStatusModifyJobPrivate()
        : m_pendingJobs( 0 )
    {
    }

    QList<Item> m_items;
    QList<KRss::StatusFlags> m_setFlags;
    QList<KRss::StatusFlags> m_clearFlags;
    int m_pendingJobs;
};

} // namespace KRss

ItemStatusModifyJob::ItemStatusModifyJob( QObject *parent ) :
    KJob( parent ), d ( new ItemStatusModifyJobPrivate )
{
}

ItemStatusModifyJob::~ItemStatusModifyJob()
{
    delete d;
}

QList<Item> ItemStatusModifyJob::items() const
{
    return d->m_items;
}

void ItemStatusModifyJob::setItems( const QList<Item>& items )
{
    d->m_items = items;
}

void ItemStatusModifyJob::setFlags( const QList<KRss::StatusFlags> &flags )
{
    d->m_setFlags = flags;
}

void ItemStatusModifyJob::clearFlags( const QList<KRss::StatusFlags> &flags )
{
    d->m_clearFlags = flags;
}

void ItemStatusModifyJob::start()
{
    QTimer::singleShot( 0, this, SLOT( doStart() ) );
}

QString ItemStatusModifyJob::errorString() const
{
    switch ( error() ) {
        case FeedLoadFailed:
        case ItemStatusModificationFailed:
            return errorText();
        default:
            return KJob::errorString();
    }

    return QString();
}

void ItemStatusModifyJob::doStart()
{
    Q_FOREACH( Item item, d->m_items ) {
        bool updated = false;

        Q_FOREACH( const KRss::StatusFlags& flag, d->m_setFlags ) {
            if ( !( item.status() & flag ) ) {
                kDebug() << "Setting" << flag << " on" << item.guid();
                item.setStatus( item.status() | flag );
                updated = true;
            }
        }

        Q_FOREACH( KRss::StatusFlags flag, d->m_clearFlags ) {
            if ( item.status() & flag ) {
                kDebug() << "Clearing" << flag << " on" << item.guid();
                item.setStatus( item.status() & ~flag );
                updated = true;
            }
        }

        if ( updated ) {
            d->m_pendingJobs++;
            ItemModifyJob * const job = new ItemModifyJob();
            job->setItem( item );
            connect( job, SIGNAL( result( KJob* ) ), this, SLOT( slotItemModificationDone( KJob* ) ) );
            job->start();
        }
    }

    checkDone();
}


void ItemStatusModifyJob::slotItemModificationDone( KJob *job )
{
    if ( job->error() ) {
        setError( ItemStatusModificationFailed );
        setErrorText( job->errorString() );
    }

    d->m_pendingJobs--;
    checkDone();
}


void ItemStatusModifyJob::checkDone()
{
    if ( d->m_pendingJobs > 0 )
        return;

    emitResult();
}

#include "itemstatusmodifyjob.moc"
