/*
    Copyright (C) 2008,2009    Dmitry Ivanov <vonami@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "persistentfeedmodifyjob.h"
#include "persistentfeed.h"
#include "persistentfeed_p.h"
#include "tagprovider.h"
#include "tagjobs.h"

#include <akonadi/collectionfetchjob.h>
#include <akonadi/collectionmodifyjob.h>
#include <KLocale>
#include <KDebug>

using namespace KRss;
using Akonadi::CollectionFetchJob;
using Akonadi::CollectionModifyJob;

namespace KRss {

class PersistentFeedModifyJobPrivate
{
public:

    explicit PersistentFeedModifyJobPrivate( const PersistentFeed *feed, PersistentFeedModifyJob *qq )
        : q( qq ), m_feed( feed )
    {
    }

    void doStart();
    void slotOriginalFeedLoaded( KJob * job );
    void slotFeedModified( KJob *job );
    void slotTagProviderRetrieved( KJob *job );
    void slotReferencesModified( KJob *job );

    PersistentFeedModifyJob * const q;
    const PersistentFeed * const m_feed;
    FeedCollection m_originalFeed;
};

} // namespace KRss

void PersistentFeedModifyJobPrivate::doStart()
{
    CollectionFetchJob *job = new CollectionFetchJob( m_feed->d->m_feedCollection, CollectionFetchJob::Base );
    QObject::connect( job, SIGNAL( result( KJob* ) ), q, SLOT( slotOriginalFeedLoaded( KJob* ) ) );
    job->start();
}

void PersistentFeedModifyJobPrivate::slotOriginalFeedLoaded( KJob *job )
{
    if ( job->error() ) {
        q->setError( PersistentFeedModifyJob::CouldNotLoadOriginalFeed );
        q->setErrorText( job->errorString() );
        kWarning() << job->errorString();
        q->emitResult();
        return;
    }

    const CollectionFetchJob* const fjob = qobject_cast<const CollectionFetchJob*>( job );
    Q_ASSERT( fjob );
    Q_ASSERT( fjob->collections().count() == 1 );
    m_originalFeed = fjob->collections().first();

    CollectionModifyJob *mjob = new CollectionModifyJob( m_feed->d->m_feedCollection );
    QObject::connect( mjob, SIGNAL( result( KJob* ) ), q, SLOT( slotFeedModified( KJob* ) ) );
    mjob->start();
}

void PersistentFeedModifyJobPrivate::slotFeedModified( KJob *job )
{
    if ( job->error() ) {
        q->setError( PersistentFeedModifyJob::CouldNotModifyFeed );
        q->setErrorText( job->errorString() );
        kWarning() << job->errorString();
        q->emitResult();
        return;
    }

    TagProviderRetrieveJob *tjob = new TagProviderRetrieveJob();
    QObject::connect( tjob, SIGNAL( result( KJob* ) ), q, SLOT( slotTagProviderRetrieved( KJob* ) ) );
    tjob->start();
}

void PersistentFeedModifyJobPrivate::slotTagProviderRetrieved( KJob *job )
{
    if ( job->error() ) {
        q->setError( PersistentFeedModifyJob::CouldNotRetrieveTagProvider );
        q->setErrorText( job->errorString() );
        kWarning() << job->errorString();
        q->emitResult();
        return;
    }

    const TagProviderRetrieveJob * const tjob = qobject_cast<const TagProviderRetrieveJob*>( job );
    Q_ASSERT( tjob );
    const TagProvider *tp = tjob->tagProvider();

    const QList<TagId> newTags = m_feed->tags();
    kDebug() << "New tags:" << newTags;
    const QList<TagId> oldTags = m_originalFeed.tags();
    kDebug() << "Old tags:" << oldTags;

    QSet<TagId> toRemove = QSet<TagId>( oldTags.toSet() ).subtract( newTags.toSet() );
    QSet<TagId> toAdd = QSet<TagId>( newTags.toSet() ).subtract( oldTags.toSet() );
    kDebug() << "Tags to remove:" << toRemove;
    kDebug() << "Tags to add:" << toAdd;

    TagModifyReferencesJob *mjob = tp->tagModifyReferencesJob();
    mjob->setReferrer( m_feed );
    mjob->setAddedTags( toAdd.toList() );
    mjob->setRemovedTags( toRemove.toList() );
    QObject::connect( mjob, SIGNAL( result( KJob* ) ), q, SLOT( slotReferencesModified( KJob* ) ) );
    mjob->start();
}

void PersistentFeedModifyJobPrivate::slotReferencesModified( KJob *job )
{
    if ( job->error() ) {
        q->setError( PersistentFeedModifyJob::CouldNotModifyReferences );
        q->setErrorText( job->errorString() );
        kWarning() << job->errorString();
        q->emitResult();
        return;
    }

    q->emitResult();
}

PersistentFeedModifyJob::PersistentFeedModifyJob( const PersistentFeed *feed, QObject *parent ) :
    KJob( parent ), d( new PersistentFeedModifyJobPrivate( feed, this ) )
{
}

PersistentFeedModifyJob::~PersistentFeedModifyJob()
{
    delete d;
}

QString PersistentFeedModifyJob::errorString() const
{
    QString result;
    switch ( error() ) {
        case KJob::NoError:
            result = i18n( "No error." );
            break;
        case PersistentFeedModifyJob::CouldNotModifyFeed:
            result = i18n( "Couldn't modify feed.\n%1", errorText() );
            break;
        case PersistentFeedModifyJob::CouldNotLoadOriginalFeed:
            result = i18n( "Couldn't load the original feed.\n%1", errorText() );
            break;
        case PersistentFeedModifyJob::CouldNotRetrieveTagProvider:
            result = i18n( "Couldn't retrieve the tag provider.\n%1", errorText() );
            break;
        case PersistentFeedModifyJob::CouldNotModifyReferences:
            result = i18n( "Couldn't modify the references inside the tag provider.\n%1", errorText() );
            break;
        default:
            result = i18n( "Unknown result code." );
            break;
    }
    return result;
}

void PersistentFeedModifyJob::start()
{
    QMetaObject::invokeMethod( this, "doStart", Qt::QueuedConnection );
}

#include "persistentfeedmodifyjob.moc"
