/*
    Copyright (C) 2008    Dmitry Ivanov <vonami@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "resource.h"
#include "krssinterface.h"
#include "importitemsjob.h"
#include "importopmljob.h"

#include <akonadi/agentinstance.h>
#include <akonadi/item.h>
#include <akonadi/monitor.h>
#include <akonadi/itemfetchscope.h>
#include <KDebug>

#include <QtCore/QString>

using Akonadi::AgentInstance;
using Akonadi::Monitor;

using namespace KRss;

namespace KRss {

class ResourcePrivate
{
public:

    explicit ResourcePrivate( const AgentInstance &instance );

    AgentInstance m_instance;
    org::kde::krss *m_interface;
    Monitor *m_monitor;
};

} // namespace KRss

ResourcePrivate::ResourcePrivate( const AgentInstance &instance )
    : m_instance( instance ), m_interface( 0 ), m_monitor( 0 )
{
}

Resource::Resource( const AgentInstance &instance, QObject *parent )
    : QObject( parent ), d( new ResourcePrivate( instance ) )
{
    d->m_interface = new org::kde::krss( "org.freedesktop.Akonadi.Agent." + d->m_instance.identifier(), "/KRss",
                                         QDBusConnection::sessionBus(), this );

    connect( d->m_interface, SIGNAL( fetchStarted( qlonglong ) ),
             this, SLOT( slotFetchStarted( qlonglong ) ) );
    connect( d->m_interface, SIGNAL( fetchPercent( qlonglong, uint ) ),
             this, SLOT( slotFetchPercent( qlonglong, uint ) ) );
    connect( d->m_interface, SIGNAL( fetchFinished( qlonglong ) ),
             this, SLOT( slotFetchFinished( qlonglong ) ) );
    connect( d->m_interface, SIGNAL( fetchFailed( qlonglong, const QString& ) ),
             this, SLOT( slotFetchFailed( qlonglong, const QString& ) ) );
    connect( d->m_interface, SIGNAL( fetchAborted( qlonglong ) ),
             this, SLOT( slotFetchAborted( qlonglong ) ) );

    // create an Akonadi::Monitor
    d->m_monitor = new Monitor( this );
    d->m_monitor->fetchCollection( false );
    d->m_monitor->itemFetchScope().fetchFullPayload();
    d->m_monitor->itemFetchScope().fetchAllAttributes();
    d->m_monitor->setResourceMonitored( d->m_instance.identifier().toLatin1(), true );
    d->m_monitor->fetchCollectionStatistics( true );

    // monitor feeds
    connect( d->m_monitor, SIGNAL( collectionAdded( const Akonadi::Collection&, const Akonadi::Collection& ) ),
             this, SLOT( slotCollectionAdded( const Akonadi::Collection&, const Akonadi::Collection& ) ) );
    connect( d->m_monitor, SIGNAL( collectionChanged( const Akonadi::Collection& ) ),
             this , SLOT( slotCollectionChanged( const Akonadi::Collection& )) );
    connect( d->m_monitor, SIGNAL( collectionRemoved( const Akonadi::Collection& ) ),
             this , SLOT( slotCollectionRemoved( const Akonadi::Collection& )) );
    connect( d->m_monitor, SIGNAL( collectionStatisticsChanged(Akonadi::Collection::Id, Akonadi::CollectionStatistics) ),
             this , SLOT( slotCollectionStatisticsChanged(Akonadi::Collection::Id, Akonadi::CollectionStatistics)) );

    // monitor items
    connect( d->m_monitor, SIGNAL( itemAdded( const Akonadi::Item&, const Akonadi::Collection& ) ),
             this, SLOT( slotItemAdded( const Akonadi::Item&, const Akonadi::Collection& ) ) );
    connect( d->m_monitor, SIGNAL( itemChanged( const Akonadi::Item&, const QSet<QByteArray>& ) ),
             this, SLOT( slotItemChanged( const Akonadi::Item&, const QSet<QByteArray>& ) ) );
    connect( d->m_monitor, SIGNAL( itemRemoved( const Akonadi::Item& ) ),
             this, SLOT( slotItemRemoved( const Akonadi::Item&) ) );
}

Resource::~Resource()
{
    delete d;
}

ImportOpmlJob* Resource::createImportOpmlJob( const KUrl& url ) const {
    ImportOpmlJob* job = new ImportOpmlJob( d->m_interface->service() );
    job->setSourceUrl( url );
    return job;
}


ImportItemsJob* Resource::createImportItemsJob( const QString& xmlUrl, const QString& sourceFile ) const {
    ImportItemsJob* job = new ImportItemsJob( d->m_interface->service() );
    job->setFeedXmlUrl( xmlUrl );
    job->setSourceFile( sourceFile );
    return job;
}

bool Resource::isValid() const
{
    return d->m_instance.isValid();
}

void Resource::addFeed( const QString &xmlUrl, const QString &subscriptionLabel ) const
{
    d->m_interface->addFeed( xmlUrl, subscriptionLabel );
}

void Resource::fetchFeed( Feed::Id id ) const
{
    d->m_interface->fetchFeed( id );
}

void Resource::abortFetch( Feed::Id id ) const
{
    d->m_interface->abortFetch( id );
}

void Resource::slotFetchStarted( qlonglong id )
{
    kDebug() << "Id:" << id;
    emit fetchStarted( id );
}

void Resource::slotFetchPercent( qlonglong id, uint percentage )
{
    kDebug() << "Id:" << id << " , percentage:" << percentage;
    emit fetchPercent( id, percentage );
}

void Resource::slotFetchFinished( qlonglong id )
{
    kDebug() << "Id:" << id;
    emit fetchFinished( id );
}

void Resource::slotFetchFailed( qlonglong id, const QString &errorMessage )
{
    kDebug() << "Id:" << id << errorMessage;
    emit fetchFailed( id, errorMessage );
}

void Resource::slotFetchAborted( qlonglong id )
{
    kDebug() << "Id:" << id;
    emit fetchAborted( id );
}

void Resource::slotCollectionAdded( const Akonadi::Collection &collection, const Akonadi::Collection &parent )
{
    kDebug() << "Added. Col:" << collection.id() << ", parent:" << parent.id();
    emit feedAdded( collection.id() );
}

void Resource::slotCollectionChanged( const Akonadi::Collection &collection )
{
    kDebug() << "Changed. Col:" << collection.id();
    emit feedChanged( collection.id() );
}

void Resource::slotCollectionRemoved( const Akonadi::Collection &collection )
{
    kDebug() << "Removed. Col:" << collection.id();
    emit feedRemoved( collection.id() );
}

void Resource::slotCollectionStatisticsChanged( Akonadi::Collection::Id id,
                                                const Akonadi::CollectionStatistics &statistics )
{
    emit statisticsChanged( id, statistics );
}

void Resource::slotItemAdded( const Akonadi::Item &item, const Akonadi::Collection &collection )
{
    kDebug() << "Added. Item:" << item.id() << " , col:" << collection.id();
    emit itemAdded( item, collection.id() );
}

void Resource::slotItemChanged( const Akonadi::Item &item, const QSet<QByteArray> &partIdentifiers )
{
    Q_UNUSED( partIdentifiers );

    kDebug() << "Changed. Item:" << item.id();
    emit itemChanged( item );
}

void Resource::slotItemRemoved( const Akonadi::Item &item )
{
    kDebug() << "Removed. Item:" << item.id();
    emit itemRemoved( item );
}

#include "resource.moc"
